package net.jieyuu.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import lombok.extern.slf4j.Slf4j;
import net.jieyuu.enums.BizCodeEnum;
import net.jieyuu.enums.CouponCategoryEnum;
import net.jieyuu.enums.CouponPublishEnum;
import net.jieyuu.enums.CouponStateEnum;
import net.jieyuu.exception.BizException;
import net.jieyuu.interceptor.LoginInterceptor;
import net.jieyuu.mapper.CouponRecordMapper;
import net.jieyuu.model.CouponDO;
import net.jieyuu.mapper.CouponMapper;
import net.jieyuu.model.CouponRecordDO;
import net.jieyuu.model.LoginUser;
import net.jieyuu.service.CouponService;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import net.jieyuu.utils.CommonUtil;
import net.jieyuu.utils.JsonData;
import net.jieyuu.vo.CouponVO;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.data.redis.core.script.DefaultRedisScript;
import org.springframework.stereotype.Service;

import java.time.Duration;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * <p>
 * 服务实现类
 * </p>
 *
 * @author jieyuu
 * @since 2024-07-15
 */
@Service
@Slf4j
public class CouponServiceImpl implements CouponService {

    @Autowired
    private CouponMapper couponMapper;

    @Autowired
    private CouponRecordMapper couponRecordMapper;

    @Autowired
    private StringRedisTemplate redisTemplate;

    @Override
    public Map<String, Object> pageCouponActivity(int page, int size) {
        Page<CouponDO> pageInfo = new Page<>(page, size);

        IPage<CouponDO> couponDOIPage = couponMapper.selectPage(pageInfo, new QueryWrapper<CouponDO>()
                .eq("publish", CouponPublishEnum.PUBLISH)
                .eq("category", CouponCategoryEnum.PROMOTION)
                .orderByAsc("create_time"));

        Map<String, Object> pageMap = new HashMap<>(3);

        pageMap.put("total_record", couponDOIPage.getTotal());
        pageMap.put("total_page", couponDOIPage.getPages());
        pageMap.put("current_data", couponDOIPage.getRecords().stream().map(obj -> beanProcess(obj)).collect(Collectors.toList()));

        return pageMap;
    }

    /**
     * 领券接口
     * 1、获取优惠券是否纯在
     * 2、校验优惠券是否可以领取: 时间 库存 超过限制
     * 3、扣减库存
     * 4、保存领券记录
     *
     * @param couponId
     * @param category
     * @return
     */
    @Override
    public JsonData addCoupon(long couponId, CouponCategoryEnum category) {
        LoginUser loginUser = LoginInterceptor.threadLocal.get();

        String uuid = CommonUtil.generateUUID();
        String lockKey = "lock:coupon:" + couponId;
        //避免锁被误删
        Boolean lockFlag = redisTemplate.opsForValue().setIfAbsent(lockKey, uuid, Duration.ofMinutes(10));

        if (lockFlag) {
            log.info("加锁成功:{}", couponId);
            try {
                // 执行业务逻辑
                CouponDO couponDO = couponMapper.selectOne(new QueryWrapper<CouponDO>()
                        .eq("id", couponId)
                        .eq("category", category.name()));
                //优惠券是否可以领取
                this.couponCheck(couponDO, loginUser.getId());

                //构建领券记录
                CouponRecordDO couponRecordDO = new CouponRecordDO();
                BeanUtils.copyProperties(couponDO, couponRecordDO);
                couponRecordDO.setCreateTime(new Date());
                couponRecordDO.setUseState(CouponStateEnum.NEW.name());
                couponRecordDO.setCouponId(couponId);
                couponRecordDO.setUserName(loginUser.getName());
                couponRecordDO.setUserId(loginUser.getId());
                couponRecordDO.setId(null);

                //扣减库存 todo
                int rows = couponMapper.reduceStock(couponId);
                if (rows == 1) {
                    //扣减库存成功才保存记录
                    couponRecordMapper.insert(couponRecordDO);
                } else {
                    log.warn("发放优惠券错误:{},用户:{}", couponDO, loginUser);
                    throw new BizException(BizCodeEnum.COUPON_NO_STOCK);
                }
            } finally {
                // 释放锁
                // 使用lua脚本保证 查询和删除的原子性
                String script = "if redis.call('get', KEYS[1]) == ARGV[1] then return redis.call('del', KEYS[1]) else return 0 end";
                Integer result = redisTemplate.execute(new DefaultRedisScript<>(script, Integer.class), Arrays.asList(lockKey), uuid);
                log.info("解锁{}", result);
            }
        } else {//加锁失败
            try {
                TimeUnit.SECONDS.sleep(1);
            } catch (InterruptedException e) {
                log.error("自旋失败");
            }
            addCoupon(couponId, category);
        }
        return JsonData.buildSuccess();
    }

    /**
     * 校验优惠券是否可以领取
     *
     * @param couponDO
     * @param id
     */
    private void couponCheck(CouponDO couponDO, Long id) {

        if (couponDO == null) {
            throw new BizException(BizCodeEnum.COUPON_NO_EXITS);
        }

        //库存不足
        if (couponDO.getStock() <= 0) {
            throw new BizException(BizCodeEnum.COUPON_NO_STOCK);
        }

        //未经发布优惠券 确认非法
        if (!couponDO.getPublish().equals(CouponPublishEnum.PUBLISH.name())) {
            throw new BizException(BizCodeEnum.COUPON_GET_FAIL);
        }

        //判断是否在领取时间
        long time = CommonUtil.getCurrentTimestamp();
        long start = couponDO.getStartTime().getTime();
        long end = couponDO.getEndTime().getTime();
        if (time < start || time > end) {
            throw new BizException(BizCodeEnum.COUPON_OUT_OF_TIME);
        }

        //用户是否超过限制
        //查询用户已经拥有的优惠券数量
        int recordNum = couponRecordMapper.selectCount(new QueryWrapper<CouponRecordDO>()
                .eq("coupon_id", couponDO.getId())
                .eq("user_id", id));

        //判断优惠券是否超过限制
        if (recordNum >= couponDO.getUserLimit()) {
            throw new BizException(BizCodeEnum.COUPON_OUT_OF_LIMIT);
        }
    }


    private CouponVO beanProcess(CouponDO couponDO) {

        CouponVO couponVO = new CouponVO();
        BeanUtils.copyProperties(couponDO, couponVO);

        return couponVO;
    }

}
